define(['EventEmitter','EventServer','http','util','querystring','fs','events'], function( EventEmitter, EventServer, http, util, querystring, fs, events ){
   
    var FileSystem = function( options ){
        
        if (!(this instanceof FileSystem)){
            return new FileSystem( options );
        }
        
        var fileSystem = this;
        var root = (options.root || require.nodeRequire.resolve('NetOS').replace(/NetOS.js$/ig, ""))+"/files";
        var netOS = options.netOS;
        
        if (!netOS){
            throw new Error("File System need a netOS reference to work properly.");
        }
        
        fileSystem.__defineGetter__("root", function(){
            return root;
        });
        
        fileSystem.__defineGetter__("netOS", function(){
            return netOS;
        });
        
        fileSystem.getFileLocation = function( path, timeout, callback ){
            if (typeof path !== 'string'){
                throw new Error("Expected the path of the file to be found.");
            }
            
            callback = Array.prototype.splice.apply(arguments, [arguments.length-1, 1])[0];
            callback = typeof callback === 'function' ? callback : function(){} ;
            timeout = typeof timeout !== 'number' ? 2000 : timeout ;

            var fs_foundFile = function(e){
                clearTimeout(t);
                callback(null, e.host);
            };
            
            var stop = function(){
                clearTimeout(t);
                netOS.removeListener("fs.foundFile"+path, fs_foundFile);
                callback(new Error("File search time out."));
            };
            
            callback.stop = stop;
            
            //Prevents memory leak.
            var t = setTimeout(stop, timeout);
            
            netOS.once("fs.foundFile"+path, fs_foundFile);
            netOS.emit("fs.hasFile", {
                path: path,
                host: netOS.host+":"+netOS.port
            });
            
        };
        
   
        fileSystem.getFileLocations = function( path, timeout, callback ){
            if (typeof path !== 'string'){
                throw new Error("Expected the path of the file to be found.");
            }
            
            callback = Array.prototype.splice.apply(arguments, [arguments.length-1, 1])[0];
            callback = typeof callback === 'function' ? callback : function(){} ;
            timeout = typeof timeout !== 'number' ? 2000 : timeout ;

            var fs_foundFile = function(e){
                callback(null, e.host);
            };
            
            var stop = function(){
                clearTimeout(t);
                netOS.removeListener("fs.foundFile"+path, fs_foundFile);
                callback(new Error("File search time out."));
            };
            
            callback.stop = stop;
            
            //Prevents memory leak.
            var t = setTimeout(stop, timeout);
            
            netOS.on("fs.foundFile"+path, fs_foundFile);
            netOS.emit("fs.hasFile", {
                path: path,
                host: netOS.host+":"+netOS.port
            });
        }
        
        if(netOS.isDrive){
            netOS.on("fs.hasFile", function(e){
                fs.realpath(root+e.path, function(err, resolvedPath){
                    if (!err){
                        netOS.emitNode(e.host, "fs.foundFile"+e.path, {
                            host: netOS.host+":"+netOS.port
                        });
                    } else {
                //console.log("Cannot find file on host: "+ netOS.host+":"+netOS.port);
                }
                });
            });
        }
    };
    FileSystem.prototype = {
        rename: function(path1, path2, callback){
            this.getFileLocation(path1, function(err, host){
                if (!err){
                    
                } else {
                    callback(err);
                }
            });
        },
        truncate: function(path1, path2, callback){},
        chmod: function(path, mode, callback){},
        stat: function(path, callback){},
        lstat: function(path, callback){},
        fstat: function(fd, callback){},
        link: function( srcpath, dstpath, callback){},
        symlink: function(linkdata, path, callback){},
        readlink: function(path, callback){},
        realPath: function(path, callback){},
        unlink: function(path, callback){},
        rmdir: function(path, callback){},
        mkdir: function(path, mode, callback){},
        readdir: function(path, callback){},
        close: function(fd, callback){},
        open: function(path, flags, mode, callback){},
        write: function(fd, buffer, offset, length, position, callback){},
        read: function(fd, buffer, offset, length, position, callback){},
        readFile: function(filename, encoding, callback){},
        watchFile: function(filename, options, listener){},
        unwatchFile: function(filename){}
    };

    var NetOS = function( options ){
        if (!(this instanceof NetOS)){
            return new NetOS();
        }
        
        options = options || {};
        
        var netOS = this;
        var host = null;
        var port = null;
        var localEmit = this.emit;
        var server = new EventServer();
        var serverFile;
        var listeners = this.listeners = {};
        var blacklistHash = {};
        var root = options.root || require.nodeRequire.resolve('NetOS').replace(/NetOS.js$/ig, "");
        var isDrive = options.isDrive || false;
        
        this.nodes = [];
        this.key = "password" || options.password;
        
        server.createServer();
        
        netOS.__defineGetter__('listeners', function(){
            return listeners;
        });
        
        netOS.__defineGetter__('host', function(){
            return host;
        });
        
        netOS.__defineGetter__('port', function(){
            return port;
        });
        
        netOS.__defineGetter__('eventServer', function(){
            return server;
        });
        
        netOS.__defineGetter__('httpServer', function(){
            return server.httpServer;
        });
        netOS.__defineGetter__('isDrive', function(){
            return isDrive;
        });
        
        var incrBlacklist = function( host ){
            if (!blacklistHash[host]){
                blacklistHash[host] = {
                    count: 0
                };
            }
            blacklistHash[host].count++;

            if (blacklistHash[host].count > 2){
                var i= netOS.nodes.indexOf(host);
                netOS.nodes.splice(i,1);
                
                //console.log("Host "+host+" removed because of failures to connect.");
                writeNodesToFile();
            }
            
        };
        
        var removeFromBlacklist = function( host ){
            if (blacklistHash[host]){
                delete blacklistHash[host];
            }
        };
        
        var writeNodesToFile = function(){
            fs.writeFile(serverFile, JSON.stringify(netOS.nodes), function(err){
                if (!err){
                //console.log("\"nodes("+host+"-"+port+").json\" updated to: "+JSON.stringify(netOS.nodes));
                } else {
                    console.log("Bigger problem, so the process is going to die.");
                    throw err;
                }
            });
        };
        
        this.emitNode = function(hostname, type){
            var args = Array.prototype.splice.apply(arguments, [2, arguments.length-2]);

            var array = hostname.split(":");
            var host = array[0];
            var port = array[1] || "80";
                       
            var node = http.createClient(port, host);
            node.on('error', function(e){
                console.log("Cannot emit \""+type+"\" event on host (using emitNode): "+hostname);
                incrBlacklist(hostname);
            });
            var request = node.request('PUT',
                ( '/'+netOS.key+'/emit?type='+type+'&args='+encodeURI(JSON.stringify(args)) ),
                {
                    'host': host
                });
            request.on("response", function(response){
                response.on('end', function(){
                    removeFromBlacklist(hostname);
                });
            });
            request.end();
        };
        
        this.emit = function(type){
            var args = Array.prototype.splice.apply(arguments, [1, arguments.length-1]);
 
            var throttle = function(x, cond){
                var nodes = netOS.nodes;
                var node;
                var host;
                var port;
                var array;
                
                if (cond(x)){
                    
                    array = nodes[x].split(":");
                    host = array[0];
                    port = array[1] || "80";
                       
                    node = http.createClient(port, host);
                    node.on('error', function(e){
                        //console.log("Cannot emit \""+type+"\" event on host: "+nodes[x]);
                        incrBlacklist(nodes[x]);
                    });
                    var request = node.request('PUT',
                        ( '/'+netOS.key+'/emit?type='+type+'&args='+encodeURI(JSON.stringify(args)) ),
                        {
                            'host': host
                        });
                    request.on("response", function(response){
                        response.on('end', function(){
                            removeFromBlacklist(nodes[x]);
                        });
                    });
                    request.end();
                        
                    setTimeout(function(){
                        throttle(x+1, cond);
                    }, 0);
                    
                } else {
                    writeNodesToFile();
                }
            };
                
            throttle(0, function(x){
                return x < netOS.nodes.length;
            });
            
        };
        
        this.listen = function( p, h, c){
            if (arguments.length < 2 || (arguments.length === 2 || typeof arguments[1] === 'function' ) ){
                throw new Error("You must specify a host, for NetOS to work properly.");
            }
            
            if (server){
                server.listen.apply(server, arguments);
                
                port = p;
                host = h;
                serverFile = root+"/nodes("+host+"-"+port+").json";
                
                fs.readFile(serverFile, function(err, data){
                    if (!err){
                        try{
                            var n = JSON.parse(data.toString());
                            if ( Object.prototype.toString.call( n ) === '[object Array]'){
                                
                                netOS.nodes = n;
                            }
                            
                        } catch (e){
                        //Nothing for now.
                        }
                    } 
                   
                    if (netOS.nodes.indexOf(netOS.host+":"+netOS.port)<0){
                        netOS.nodes.push(netOS.host+":"+netOS.port);
                    }
                    
                });
                
                
            } else {
                throw new Error("No server to listen with yet.");
            }
            return server;
        };
        
        this.close = function(){
            if (server){
                server.close.apply(server, arguments);
            } else {
                throw new Error("No server to stop.");
            }
            return server;
        };
        
        this.fs = new FileSystem({
            netOS: netOS,
            root: root
        });
        
        server.on("noExtension", function(e){
            if (e.url.path.indexOf(netOS.key+'/emit') === 0 ){
                if (e.url.query){
                    
                    var params = querystring.parse(e.url.query);
                    var type = params.type;
                    
                    var args = JSON.parse(params.args || "[]");

                    if (type){
                        args.unshift(type);
                        localEmit.apply(netOS, args);
                        return false;
                    }
                }
            }
        });
        
        server.on("noExtension", function(e){
            if (e.url.path.indexOf(netOS.key+'/invoke') === 0 ){
                var params = querystring.parse(e.url.query);
                var fn = eval(params.fn);
                    
                if (fn){
                    fn.apply(null, [e]);
                    return false;
                }
            }
        });
        
        server.on("noExtension", function(e){
            if (e.url.path.indexOf(netOS.key+'/nodes') === 0 ){
                if (e.url.query){
                    var params = querystring.parse(e.url.query);
                    
                    if (params.host){
                        if (e.request.method === 'PUT'){
                            if (netOS.nodes.indexOf(params.host)<0){
                                netOS.nodes.push(params.host);
                            }
                            e.response.writeHead(200,{
                                "Content-Type":"application/json"
                            });
                            e.response.end(JSON.stringify(netOS.nodes));
                            netOS.emit("nodesChange",netOS.nodes);
                            return false;
                        } else if (e.request.method === 'DELETE'){
                            var i = netOS.nodes.indexOf(params.host);
                            if ( i > -1 ){
                                netOS.nodes.splice(i,1);
                            }
                            e.response.writeHead(200,{
                                "Content-Type":"application/json"
                            });
                            e.response.end(JSON.stringify(netOS.nodes));
                            return false;
                        }
                    } 
                }
                
                if (e.request.method === 'GET'){
                    e.response.writeHead(200,{
                        "Content-Type":"application/json"
                    });
                    e.response.end(JSON.stringify(netOS.nodes));
                    return false;
                }
            }
        });
        
        netOS.on("nodesChange", function( n ){
            if (Object.prototype.toString.call( n ) === '[object Array]'){
                //Throttling to not jam a node process.
                var throttle = function(x, cond){
                    var nodes = netOS.nodes;
                    if (cond(x)){
                        
                        if (nodes.indexOf(n[x])<0){
                            nodes.push(n[x]);
                        }
                        
                        setTimeout(function(){
                            throttle(x+1, cond);
                        }, 0);
                    } else {
                        writeNodesToFile();
                    }
                };
                
                throttle(0, function(x){
                    return x < n.length;
                });
                
            }
        });
        
    };
    
    NetOS.prototype = new EventEmitter();
    
    return NetOS;
});